--

separate (Prime_Numbers)

task body Sieve_Task_Type is

   Input_Buffer : Buffers.Buffer_Type renames Chain_Member.Buffer;

   Next_Sieve_Chain_Member : access Sieve_Chain_Member_Type;
   --
   --  The Output_Buffer is Next_Sieve_Chain_Member.Buffer

   Num      : Natural;
   My_Prime : Natural;

   -- -- --

   procedure Do_Output;
   --
   --  Uses global data: Last_Prime; Uses local data: My_Prime

   procedure Do_Output is
   begin
      if (My_Prime - Last_Prime) = 2 then

         Printing : declare
            Str : String :=
              ">> PrimePair: " & Positive'Image (Last_Prime)
            --
             & " | " & Positive'Image (My_Prime);

         begin
            Printer_Task.Print (Str);
         end Printing;

         Nb_Of_Prime_Pairs := Natural'Succ (Nb_Of_Prime_Pairs);

      end if;

      Last_Prime := My_Prime;

   end Do_Output;

   -- -- --

   procedure Handle_The_End (N : Natural);
   pragma Inline (Handle_The_End);

   procedure Handle_The_End (N : Natural) is
   begin
      if N = 0 then
         raise Prime_Numbers.The_End;
      end if;
   end Handle_The_End;

   -- -- --

   procedure Get_Number (N : out Natural);
   --
   -- Uses Input_Buffer

   procedure Get_Number (N : out Natural) is
   begin
      Buffers.Get (Input_Buffer, N);
      Handle_The_End (N);
   end Get_Number;

   -- -- --

   function Remainder_On_Division (N : Natural) return Boolean;
   pragma Inline(Remainder_On_Division);
   --
   -- Uses My_Prime

   function Remainder_On_Division (N : Natural) return Boolean is
   begin
      return N rem My_Prime /= 0;
   end Remainder_On_Division;

-- -- --

begin

   Get_Number (Num);

   My_Prime := Num;
   --
   --  The first number is my prime number

   Do_Output;

   -- -- --

   loop
      Get_Number (Num);

      exit when Remainder_On_Division (Num);
   end loop;

   -- -- --

   -- Received the next prime number,
   -- so create a sieve task and a buffer
   -- to output the number to the new task

   Next_Sieve_Chain_Member := Get_New_Sieve_Chain_Member;

   Buffers.Put (Next_Sieve_Chain_Member.Buffer, Num);

   -- -- --

   loop
      Get_Number (Num);

      if Remainder_On_Division (Num) then

         Buffers.Put (Next_Sieve_Chain_Member.Buffer, Num);
      end if;

   end loop;

   -- -- --

exception

   when Event : Prime_Numbers.The_End =>

      if Next_Sieve_Chain_Member /= null then

         Buffers.Put (Next_Sieve_Chain_Member.Buffer, 0);
         --
         -- Put 0 to indicate the end

      end if;

      delay 1.0;

      if not Done then
         Done := True;
         --
         -- Global atomic data

         Printer_Task.Print ("#> ---");
         Printer_Task.Print ("#> Number of prime pairs: " & Natural'Image (Nb_Of_Prime_Pairs));
         Printer_Task.Print ("## Enter CR to finish ...");
      end if;

   when Event : others =>
      Printer_Task.Print ("#! Sieve_Task_Type: EXCEPTION");
      Printer_Task.Print (Exceptions.Exception_Information (Event));

end Sieve_Task_Type;
