--

with Bounded_Buffers;
with Printer;

with Ada.Exceptions;
with System.Multiprocessors;

package body Prime_Numbers is
   use Ada;
   use Printer;

   -- -- --

   Done : Boolean := False;
   pragma Atomic (Done);

   Last_Prime : Positive := 2;
   pragma Atomic (Last_Prime);

   Nb_Of_Prime_Pairs : Natural := 0;
   pragma Atomic (Nb_Of_Prime_Pairs);

   Task_Nb : Integer := -1;
   pragma Atomic (Task_Nb);

   My_Number_Of_Cpus : Natural := Natural (System.Multiprocessors.Number_Of_CPUs);
   pragma Atomic (My_Number_Of_Cpus);

   function Set_Cpu return System.Multiprocessors.CPU_Range;

   -- -- --

   package Buffers is new Bounded_Buffers (Item_Type => Natural);

   subtype Source_Buffer_Type is
     Buffers.Buffer_Type (Buffer_Size => Source_Buffer_Size, Ceiling => Ceiling_Priority);

   -- -- --

   task Source_Task is
      pragma Priority (Source_Task_Priority);

      entry Start;
   end Source_Task;

   -- -- --

   type Sieve_Chain_Member_Type;

   task type Sieve_Task_Type (Chain_Member : not null access Sieve_Chain_Member_Type) with
      Cpu      => Set_Cpu,
      Priority => Sieve_Task_Priority;

   type Sieve_Chain_Member_Type
     (Buffer_Size : Positive;
      Ceiling     : System.Priority) is limited record
      --
      Buffer     : aliased Buffers.Buffer_Type (Buffer_Size => Buffer_Size, Ceiling => Ceiling);
      Sieve_Task : Sieve_Task_Type (Chain_Member => Sieve_Chain_Member_Type'Access);
   end record;

   -- -- --

   function Get_New_Sieve_Chain_Member
     (Buffer_Size : Positive        := Sieve_Buffer_Size;
      Ceiling     : System.Priority := Ceiling_Priority) return access Sieve_Chain_Member_Type;

   -- -- --
   -- -- --

   task body Sieve_Task_Type is separate;

   -- -- --

   task body Source_Task is
      Limit : constant Positive := Limit_Integer;

      Num : Natural;

      Next_Sieve_Chain_Member : access Sieve_Chain_Member_Type;
   begin
      accept Start;

      Num := 3;

      Next_Sieve_Chain_Member := Get_New_Sieve_Chain_Member (Buffer_Size => Source_Buffer_Size);

      while Num <= Limit loop

         Buffers.Put (Next_Sieve_Chain_Member.Buffer, Num);

         Num := Num + 2;
      end loop;

      delay 1.0;

      Num := 0;
      Buffers.Put (Next_Sieve_Chain_Member.Buffer, Num);

      delay 1.0;

   end Source_Task;

   -- -- --
   -- -- --

   function Set_Cpu return System.Multiprocessors.CPU_Range is
      use System.Multiprocessors;
   begin
      Task_Nb := Integer'Succ (Task_Nb);

      return CPU_Range (Task_Nb mod My_Number_Of_Cpus + 1);

   exception
      when Constraint_Error =>
         Task_Nb := 0;

         return 1;
   end Set_Cpu;

   -- -- --

   function Get_New_Sieve_Chain_Member
     (Buffer_Size : Positive        := Sieve_Buffer_Size;
      Ceiling     : System.Priority := Ceiling_Priority) return access Sieve_Chain_Member_Type is
   begin
      return new Sieve_Chain_Member_Type (Buffer_Size, Ceiling);
   end Get_New_Sieve_Chain_Member;

   -- -- --
   -- -- --

   procedure Start is
   begin
      Source_Task.Start;
   end Start;

end Prime_Numbers;
